# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import logging
import typing as t
from socket import (
    AddressFamily,
    AF_INET,
    AF_INET6,
    getservbyname,
)


if t.TYPE_CHECKING:
    import typing_extensions as te


log = logging.getLogger("neo4j")


_T = t.TypeVar("_T")


if t.TYPE_CHECKING:

    class _WithPeerName(te.Protocol):
        def getpeername(self) -> tuple: ...


assert type(tuple) is type


class _AddressMeta(type):

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        cls._ipv4_cls = None
        cls._ipv6_cls = None

    def _subclass_by_family(self, family):
        subclasses = [
            sc for sc in self.__subclasses__()
            if (sc.__module__ == self.__module__
                and getattr(sc, "family", None) == family)
        ]
        if len(subclasses) != 1:
            raise ValueError(
                "Class {} needs exactly one direct subclass with attribute "
                "`family == {}` within this module. "
                "Found: {}".format(self, family, subclasses)
            )
        return subclasses[0]

    @property
    def ipv4_cls(self):
        if self._ipv4_cls is None:
            self._ipv4_cls = self._subclass_by_family(AF_INET)
        return self._ipv4_cls

    @property
    def ipv6_cls(self):
        if self._ipv6_cls is None:
            self._ipv6_cls = self._subclass_by_family(AF_INET6)
        return self._ipv6_cls


class Address(tuple, metaclass=_AddressMeta):

    @classmethod
    def from_socket(
        cls: t.Type[_T_Address],
        socket: _WithPeerName
    ) -> _T_Address:
        address = socket.getpeername()
        return cls(address)

    @classmethod
    def parse(
        cls: t.Type[_T_Address],
        s: str,
        default_host: str = None,
        default_port: int = None
    ) -> _T_Address:
        if not isinstance(s, str):
            raise TypeError("Address.parse requires a string argument")
        if s.startswith("["):
            # IPv6
            port: t.Union[str, int]
            host, _, port = s[1:].rpartition("]")
            port = port.lstrip(":")
            try:
                port = int(port)
            except (TypeError, ValueError):
                pass
            return cls((host or default_host or "localhost",
                        port or default_port or 0, 0, 0))
        else:
            # IPv4
            host, _, port = s.partition(":")
            try:
                port = int(port)
            except (TypeError, ValueError):
                pass
            return cls((host or default_host or "localhost",
                        port or default_port or 0))

    @classmethod
    def parse_list(
        cls: t.Type[_T_Address],
        *s: str,
        default_host: str = None,
        default_port: int = None
    ) -> t.List[_T_Address]:
        """ Parse a string containing one or more socket addresses, each
        separated by whitespace.
        """
        if not all(isinstance(s0, str) for s0 in s):
            raise TypeError("Address.parse_list requires a string argument")
        return [cls.parse(a, default_host, default_port)
                for a in " ".join(s).split()]

    def __new__(cls, iterable: t.Collection) -> Address:
        if isinstance(iterable, cls):
            return iterable
        n_parts = len(iterable)
        inst = tuple.__new__(cls, iterable)
        if n_parts == 2:
            inst.__class__ = cls.ipv4_cls
        elif n_parts == 4:
            inst.__class__ = cls.ipv6_cls
        else:
            raise ValueError("Addresses must consist of either "
                             "two parts (IPv4) or four parts (IPv6)")
        return inst

    #: Address family (AF_INET or AF_INET6)
    family: t.Optional[AddressFamily] = None

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, tuple(self))

    @property
    def host_name(self) -> str:
        return self[0]

    @property
    def host(self) -> str:
        return self[0]

    @property
    def port(self) -> int:
        return self[1]

    @property
    def unresolved(self) -> Address:
        return self

    @property
    def port_number(self) -> int:
        try:
            return getservbyname(self[1])
        except (OSError, TypeError):
            # OSError: service/proto not found
            # TypeError: getservbyname() argument 1 must be str, not X
            try:
                return int(self[1])
            except (TypeError, ValueError) as e:
                raise type(e)("Unknown port value %r" % self[1])


_T_Address = t.TypeVar("_T_Address", bound=Address)


class IPv4Address(Address):

    family = AF_INET

    def __str__(self) -> str:
        return "{}:{}".format(*self)


class IPv6Address(Address):

    family = AF_INET6

    def __str__(self) -> str:
        return "[{}]:{}".format(*self)


class ResolvedAddress(Address):

    _host_name: str

    @property
    def host_name(self) -> str:
        return self._host_name

    @property
    def unresolved(self) -> Address:
        return super().__new__(Address, (self._host_name, *self[1:]))

    def __new__(cls, iterable, *, host_name: str) -> ResolvedAddress:
        new = super().__new__(cls, iterable)
        new = t.cast(ResolvedAddress, new)
        new._host_name = host_name
        return new


class ResolvedIPv4Address(IPv4Address, ResolvedAddress):
    pass


class ResolvedIPv6Address(IPv6Address, ResolvedAddress):
    pass



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import typing as t
from logging import (
    CRITICAL,
    DEBUG,
    ERROR,
    Formatter,
    getLogger,
    INFO,
    StreamHandler,
    WARNING,
)
from sys import stderr


__all__ = [
    "Watcher",
    "watch"
]


class ColourFormatter(Formatter):
    """ Colour formatter for pretty log output.
    """

    def format(self, record):
        s = super(ColourFormatter, self).format(record)
        if record.levelno == CRITICAL:
            return "\x1b[31;1m%s\x1b[0m" % s  # bright red
        elif record.levelno == ERROR:
            return "\x1b[33;1m%s\x1b[0m" % s  # bright yellow
        elif record.levelno == WARNING:
            return "\x1b[33m%s\x1b[0m" % s    # yellow
        elif record.levelno == INFO:
            return "\x1b[37m%s\x1b[0m" % s    # white
        elif record.levelno == DEBUG:
            return "\x1b[36m%s\x1b[0m" % s    # cyan
        else:
            return s


class Watcher:
    """Log watcher for easier logging setup.

    Example::

        from neo4j.debug import Watcher

        with Watcher("neo4j"):
            # DEBUG logging to stderr enabled within this context
            ...  # do something

    .. note:: The Watcher class is not thread-safe. Having Watchers in multiple
        threads can lead to duplicate log messages as the context manager will
        enable logging for all threads.

    :param logger_names: Names of loggers to watch.
    :param default_level: Default minimum log level to show.
        The level can be overridden by setting the level a level when calling
        :meth:`.watch`.
    :param default_out: Default output stream for all loggers.
        The level can be overridden by setting the level a level when calling
        :meth:`.watch`.
    :param colour: Whether the log levels should be indicated with ANSI colour
        codes.
    """

    def __init__(
        self,
        *logger_names: str,
        default_level: int = DEBUG,
        default_out: t.TextIO = stderr,
        colour: bool = False
    ) -> None:
        super(Watcher, self).__init__()
        self.logger_names = logger_names
        self._loggers = [getLogger(name) for name in self.logger_names]
        self.default_level = default_level
        self.default_out = default_out
        self._handlers: t.Dict[str, StreamHandler] = {}

        format_ = "%(threadName)s(%(thread)d) %(asctime)s  %(message)s"
        if not colour:
            format_ = "[%(levelname)-8s] " + format_
        formatter_cls = ColourFormatter if colour else Formatter
        self.formatter = formatter_cls(format_)

    def __enter__(self) -> Watcher:
        """Enable logging for all loggers."""
        self.watch()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Disable logging for all loggers."""
        self.stop()

    def watch(self, level: int = None, out: t.TextIO = None):
        """Enable logging for all loggers.

        :param level: Minimum log level to show.
            If :const:`None`, the ``default_level`` is used.
        :param out: Output stream for all loggers.
            If :const:`None`, the ``default_out`` is used.
        """
        if level is None:
            level = self.default_level
        if out is None:
            out = self.default_out
        self.stop()
        handler = StreamHandler(out)
        handler.setFormatter(self.formatter)
        for logger in self. _loggers:
            self._handlers[logger.name] = handler
            logger.addHandler(handler)
            logger.setLevel(level)

    def stop(self) -> None:
        """Disable logging for all loggers."""
        for logger in self._loggers:
            try:
                logger.removeHandler(self._handlers.pop(logger.name))
            except KeyError:
                pass


def watch(
    *logger_names: str,
    level: int = DEBUG,
    out: t.TextIO = stderr,
    colour: bool = False
) -> Watcher:
    """Quick wrapper for using  :class:`.Watcher`.

    Create a Wathcer with the given configuration, enable watching and return
    it.

    Example::

        from neo4j.debug import watch

        watch("neo4j")
        # from now on, DEBUG logging to stderr is enabled in the driver

    :param logger_names: Names of loggers to watch.
    :type logger_names: str
    :param level: see ``default_level`` of :class:`.Watcher`.
    :type level: int
    :param out: see ``default_out`` of :class:`.Watcher`.
    :type out: stream or file-like object
    :param colour: see ``colour`` of :class:`.Watcher`.
    :type colour: bool

    :return: Watcher instance
    :rtype: :class:`.Watcher`
    """
    watcher = Watcher(*logger_names, colour=colour, default_level=level,
                      default_out=out)
    watcher.watch()
    return watcher



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Request(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._seen_keys = set()

    def __getitem__(self, item):
        self._seen_keys.add(item)
        return super().__getitem__(item)

    def get(self, item, default=None):
        self._seen_keys.add(item)
        return super(Request, self).get(item, default)

    def mark_all_as_read(self, recursive=False):
        self._seen_keys = set(self.keys())
        if recursive:
            for val in self.values():
                if isinstance(val, Request):
                    val.mark_all_as_read(recursive=True)

    def mark_item_as_read(self, item, recursive=False):
        self._seen_keys.add(item)
        if recursive:
            value = super().__getitem__(item)
            if isinstance(value, Request):
                value.mark_all_as_read(recursive=True)

    def mark_item_as_read_if_equals(self, item, value, recursive=False):
        if super().__getitem__(item) == value:
            self.mark_item_as_read(item, recursive=recursive)

    @property
    def unseen_keys(self):
        assert not any(isinstance(v, dict) and not isinstance(v, Request)
                       for v in self.values())
        unseen = set(self.keys()) - self._seen_keys
        for k, v in self.items():
            if isinstance(v, Request):
                unseen.update(k + "." + u for u in v.unseen_keys)
        return unseen

    @property
    def seen_all_keys(self):
        return not self.unseen_keys



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
from socketserver import (
    StreamRequestHandler,
    TCPServer,
)

from ._async.backend import AsyncBackend
from ._sync.backend import Backend


class Server(TCPServer):
    allow_reuse_address = True

    def __init__(self, address):
        class Handler(StreamRequestHandler):
            def handle(self):
                backend = Backend(self.rfile, self.wfile)
                try:
                    while backend.process_request():
                        pass
                finally:
                    backend.close()
                print("Disconnected")
        super(Server, self).__init__(address, Handler)


class AsyncServer:
    def __init__(self, address):
        self._address = address
        self._server = None

    @staticmethod
    async def _handler(reader, writer):
        backend = AsyncBackend(reader, writer)
        try:
            while await backend.process_request():
                pass
        finally:
            writer.close()
            await backend.close()
        print("Disconnected")

    async def start(self):
        self._server = await asyncio.start_server(
            self._handler, host=self._address[0], port=self._address[1],
            limit=float("inf")  # this is dirty but works (for now)
        )

    async def serve_forever(self):
        if not self._server:
            raise RuntimeError("Server not started")
        await self._server.serve_forever()

    def stop(self):
        if not self._server:
            raise RuntimeError("Try starting the server before stopping it ;)")
        self._server.close()



#!/usr/bin/env python

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
import sys
import warnings

from .server import (
    AsyncServer,
    Server,
)


def sync_main():
    server = Server(("0.0.0.0", 9876))
    while True:
        server.handle_request()


def async_main():
    async def main():
        server = AsyncServer(("0.0.0.0", 9876))
        await server.start()
        try:
            await server.serve_forever()
        finally:
            server.stop()

    asyncio.run(main())


if __name__ == "__main__":
    warnings.simplefilter("error")
    if len(sys.argv) == 2 and sys.argv[1].lower().strip() == "async":
        async_main()
    else:
        sync_main()



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import typing as t
from collections import defaultdict

from .._async_compat.concurrency import CooperativeLock
from .._async_compat.util import Util
from ..api import (
    BookmarkManager,
    Bookmarks,
)


T_BmSupplier = t.Callable[[t.Optional[str]],
                          t.Union[Bookmarks, t.Union[Bookmarks]]]
T_BmConsumer = t.Callable[[str, Bookmarks], t.Union[None, t.Union[None]]]


def _bookmarks_to_set(
    bookmarks: t.Union[Bookmarks, t.Iterable[str]]
) -> t.Set[str]:
    if isinstance(bookmarks, Bookmarks):
        return set(bookmarks.raw_values)
    return set(map(str, bookmarks))


class Neo4jBookmarkManager(BookmarkManager):
    def __init__(
        self,
        initial_bookmarks: t.Mapping[str, t.Union[Bookmarks,
                                                  t.Iterable[str]]] = None,
        bookmarks_supplier: T_BmSupplier = None,
        bookmarks_consumer: T_BmConsumer = None
    ) -> None:
        super().__init__()
        self._bookmarks_supplier = bookmarks_supplier
        self._bookmarks_consumer = bookmarks_consumer
        if initial_bookmarks is None:
            initial_bookmarks = {}
        self._bookmarks = defaultdict(
            set, ((k, _bookmarks_to_set(v))
                  for k, v in initial_bookmarks.items())
        )
        self._lock = CooperativeLock()

    def update_bookmarks(
        self, database: str, previous_bookmarks: t.Collection[str],
        new_bookmarks: t.Collection[str]
    ) -> None:
        if not new_bookmarks:
            return
        with self._lock:
            curr_bms = self._bookmarks[database]
            curr_bms.difference_update(previous_bookmarks)
            curr_bms.update(new_bookmarks)
            if self._bookmarks_consumer:
                curr_bms_snapshot = Bookmarks.from_raw_values(curr_bms)
        if self._bookmarks_consumer:
            Util.callback(
                self._bookmarks_consumer, database, curr_bms_snapshot
            )

    def get_bookmarks(self, database: str) -> t.Set[str]:
        with self._lock:
            bms = set(self._bookmarks[database])
        if self._bookmarks_supplier:
            extra_bms = Util.callback(
                self._bookmarks_supplier, database
            )
            bms.update(extra_bms.raw_values)
        return bms

    def get_all_bookmarks(self) -> t.Set[str]:
        bms: t.Set[str] = set()
        with self._lock:
            for database in self._bookmarks.keys():
                bms.update(self._bookmarks[database])
        if self._bookmarks_supplier:
            extra_bms = Util.callback(
                self._bookmarks_supplier, None
            )
            bms.update(extra_bms.raw_values)
        return bms

    def forget(self, databases: t.Iterable[str]) -> None:
        with self._lock:
            for database in databases:
                self._bookmarks.pop(database, None)



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Functions to decide whether to run a subtest or not.

They take the subtest parameters as arguments and return
 - a string with describing the reason why the subtest should be skipped
 - None if the subtest should be run
"""


import pytz

from . import fromtestkit


def tz_id(**params):
    # We could do this automatically, but with an explicit black list we
    # make sure we know what we test and what we don't.
    # if params["tz_id"] not in pytz.common_timezones_set:
    #     return (
    #         "timezone id %s is not supported by the system" % params["tz_id"]
    #     )

    if params["tz_id"] in {
        "SystemV/AST4",
        "SystemV/AST4ADT",
        "SystemV/CST6",
        "SystemV/CST6CDT",
        "SystemV/EST5",
        "SystemV/EST5EDT",
        "SystemV/HST10",
        "SystemV/MST7",
        "SystemV/MST7MDT",
        "SystemV/PST8",
        "SystemV/PST8PDT",
        "SystemV/YST9",
        "SystemV/YST9YDT",
    }:
        return (
            "timezone id %s is not supported by the system" % params["tz_id"]
        )


def dt_conversion(**params):
    dt = params["dt"]
    try:
        fromtestkit.to_param(dt)
    except (pytz.UnknownTimeZoneError, ValueError) as e:
        return "cannot create desired dt %s: %r" % (dt, e)
